#include <errno.h>
#include <pthread.h>    /* pthread_* */
#include <pthread_np.h> /* pthread w/ affinity */
#include <signal.h>     /* signal */
#include <stdlib.h>
#include <stdio.h>
#include <inttypes.h>   /* PRI* macros */
#include <string.h>     /* strcmp */
#include <fcntl.h>      /* open */
#include <unistd.h>     /* close */
#include <ifaddrs.h>    /* getifaddrs */

#include <sys/mman.h>   /* PROT_* */
#include <sys/ioctl.h>  /* ioctl */
#include <sys/poll.h>
#include <sys/socket.h> /* sockaddr.. */
#include <arpa/inet.h>  /* ntohs */
#include <sys/param.h>
#include <sys/cpuset.h> /* cpu_set */
#include <sys/sysctl.h> /* sysctl */
#include <sys/time.h>   /* timersub */

#include <net/ethernet.h>
#include <net/if.h>     /* ifreq */
#include <net/if_dl.h>  /* LLADDR */

#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/udp.h>

#include <net/netmap.h>
#include <net/netmap_user.h>
#include <pcap/pcap.h>

#include <curses.h>
#include <term.h>

/* debug support */
#define D(format, ...)                          \
	wprintw(WIN_Log, "%s [%d] " format "\n",         \
		__FUNCTION__, __LINE__, ##__VA_ARGS__); wrefresh(WIN_Log);;

static uint64_t freq = 2793376516; // machdep.tsc_freq
static unsigned long long Speed= 10000000000; // 
static unsigned long long LoopCnt = -1;
static unsigned SendMode = 1; // parallel sending


static __inline__ unsigned long long rdtsc(void)
{
	unsigned a, d;
	__asm__ volatile("rdtsc" : "=a" (a), "=d" (d));
	return a | (((unsigned long long)d) << 32);;
}

struct targ {
	pthread_t thread;
	int fd;
	struct nmreq nmr;
	struct netmap_if *nifp;
	uint64_t p_count, p_bytes, run_count;
	uint64_t sample_bytes, sample_count;
	int nfiles, nsamplefiles;
	int me, done;
	int batch_size;
	unsigned SendMode;
	unsigned long long LoopCnt;
	uint8_t dst_mac[6];
	uint8_t src_mac[6];
};



static void
usage(const char *progname)
{
	fprintf(stderr, "Usage: %s -i interface [-s speed] [-l loopcount] [-p=0|1] pcap1 [pcap2] ... [pcapN] \n", progname);
	exit(0);
}

struct pcap_ring_t {
	struct pcap_ring_t *next;
	unsigned int len;
	u_char buf[0];
};
	
static struct pcap_ring_t **pcap_ring;
static struct pcap_ring_t **pcap_sample_ring;

static void *
send_thread(void *data)
{
	struct targ *targ = (struct targ *) data;
	struct pollfd fds[1];
	struct netmap_if *nifp = targ->nifp;
	struct netmap_ring *txring, *rxring;
	int ifile, isfile;
	/* setup poll(2) machanism. */
	memset(fds, 0, sizeof(fds));
	fds[0].fd = targ->fd;
	fds[0].events = (POLLOUT);

	/* main loop.*/
	struct pcap_ring_t **c_pcap = malloc( targ->nfiles*sizeof(*pcap_ring));
	for(ifile=0; ifile<targ->nfiles; ifile++) c_pcap[ifile] = pcap_ring[ifile];
	struct pcap_ring_t **cs_pcap = malloc( targ->nsamplefiles*sizeof(*pcap_sample_ring));
	for(isfile=0; isfile<targ->nsamplefiles; isfile++) cs_pcap[isfile] = pcap_sample_ring[isfile];
	ifile=0; isfile=0;
	int batch_size = targ->batch_size, batch = 0;
	double bpt_total, bpt_sample;
	bpt_total = freq/(Speed/8.0);
	unsigned long long tsc_sample, tsc = rdtsc();
	if(pcap_sample_ring[0]) {
		bpt_sample = freq/(500000000/8.0);
		tsc_sample = tsc;
	} else {
		bpt_sample = 0;
		tsc_sample = 0;
	}
	int Set = 0;
	while(1){
		if (poll(fds, 1, 2000) <= 0) { // wait for available room in the send queue(s)
			fprintf(stderr,"poll error/timeout on queue %d\n", targ->me);
			goto quit;
		}
		if(targ->run_count >= targ->LoopCnt) break;
		unsigned int sent=0; batch=batch_size;
		rxring = NETMAP_RXRING(nifp, targ->nmr.nr_rx_rings); // host ring
		for (int i = 0; i < targ->nmr.nr_tx_rings; i++) {
			int m;
			txring = NETMAP_TXRING(nifp, i);
			int count = txring->avail;
			if (count == 0) continue;
			if(count>batch) count=batch;
			batch -= count;
			// m = send_packets(txring, &targ->pkt, targ->g->pkt_size, limit, fill_all);
			// send_packets(struct netmap_ring *ring, struct pkt *pkt, int size, u_int count, int fill_all)
			u_int cur = txring->cur;
			for (m = 0; m < count; m++) {
				register unsigned bytes;
				register struct pcap_ring_t *pcap;
				struct netmap_slot *slot = &txring->slot[cur];
				if(Set) {
					if(tsc_sample > rdtsc()) Set=0;
					else pcap = cs_pcap[isfile];
				}
				if(Set == 0) {
					pcap = c_pcap[ifile];
				}
				memcpy(NETMAP_BUF(txring, slot->buf_idx), pcap->buf, pcap->len);
				bytes = pcap->len+4;
				sent += bytes;
				targ->p_count++; targ->p_bytes += bytes;
				slot->len = pcap->len;
				if(Set == 0) {
					c_pcap[ifile] = pcap = pcap->next;
					if(targ->SendMode || pcap == pcap_ring[ifile])
						if(++ifile >= targ->nfiles) {
							ifile = 0;
							if(pcap == pcap_ring[ifile]) {
								targ->run_count++;
								if(targ->run_count >= targ->LoopCnt) count = m;
							}
						}
				} else {
					targ->sample_count++;
					targ->sample_bytes += bytes;
					if(pcap->next == pcap_sample_ring[isfile]) cs_pcap[isfile] = NULL; // last packet sent, clear
					else cs_pcap[isfile] = pcap->next;
					register int i;
					for(i = isfile+1; i != isfile; i++) {
						if(i>= targ->nsamplefiles) { i = -1; continue; }
						if(cs_pcap[i] == NULL) continue;
					}
					if(cs_pcap[i] == NULL) { // all packet from all files sent
						tsc_sample = 0;
						Set = 0;
					} else {
						isfile = i;
						tsc_sample += bytes*bpt_sample;
					}
				}
				if(tsc_sample && ++Set > 1) Set = 0;
				//if (m == count - 1) slot->flags |= NS_REPORT;
				cur = NETMAP_RING_NEXT(txring, cur);
			}
			txring->avail -= m;
			txring->cur = cur;
			// insert packets from host
			if(rxring->avail && txring->avail) {
				uint32_t pkt;
				pkt = txring->slot[txring->cur].buf_idx;
				txring->slot[txring->cur].buf_idx =  rxring->slot[rxring->cur].buf_idx;
				rxring->slot[rxring->cur].buf_idx = pkt;
				txring->slot[txring->cur].len = rxring->slot[rxring->cur].len;
				txring->slot[txring->cur].flags |= NS_BUF_CHANGED;
				rxring->slot[rxring->cur].buf_idx |= NS_BUF_CHANGED;
				txring->cur = NETMAP_RING_NEXT(txring, txring->cur);
				rxring->cur = NETMAP_RING_NEXT(rxring, rxring->cur);
				rxring->avail--;
				txring->avail--;
			}
		}
		// send packets from interface to host
		txring = NETMAP_TXRING(nifp, targ->nmr.nr_tx_rings); // host ring
		if(txring->avail) 
			for (int i = 0; i < targ->nmr.nr_rx_rings; i++) {
				if (txring->avail == 0) break;
				rxring = NETMAP_RXRING(nifp, i);
				while(rxring->avail && txring->avail){
					uint32_t pkt;
					pkt = txring->slot[txring->cur].buf_idx;
					txring->slot[txring->cur].buf_idx =  rxring->slot[rxring->cur].buf_idx;
					rxring->slot[rxring->cur].buf_idx = pkt;
					txring->slot[txring->cur].len = rxring->slot[rxring->cur].len;
					txring->slot[txring->cur].flags |= NS_BUF_CHANGED;
					rxring->slot[rxring->cur].buf_idx |= NS_BUF_CHANGED;
					txring->cur = NETMAP_RING_NEXT(txring, txring->cur);
					rxring->cur = NETMAP_RING_NEXT(rxring, rxring->cur);
					rxring->avail--;
					txring->avail--;
				}
			}
		//
		tsc += sent*bpt_total;
		while(tsc>rdtsc());
	}
	/* final part: wait all the TX queues to be empty. */
	for (int i = 0; i < targ->nmr.nr_tx_rings; i++) {
		txring = NETMAP_TXRING(nifp, i);
		while (!NETMAP_TX_RING_EMPTY(txring)) {
			ioctl(fds[0].fd, NIOCTXSYNC, NULL);
			usleep(1); /* wait 1 tick */
		}
	}
quit:
	/* reset the ``used`` flag. */
	targ->done = 1;

	return (NULL);
}

struct pcap_ring_t *
load_pcap_file(char *name)
{
		static struct ether_header eth_hdr = { .ether_dhost = {0x00,0x1b,0x21,0x7e,0x74,0x16},
						.ether_shost = {0x00,0x1b,0x21,0x62,0xa3,0xe2},
						.ether_type  = 0x0008 };
		char errbuf[PCAP_ERRBUF_SIZE];
		pcap_t *pcap_file = pcap_open_offline(name, errbuf);
		if(pcap_file == NULL) return NULL;
		struct pcap_pkthdr pkt_header;
		const u_char *pcap_buf;
		struct pcap_ring_t *c_ring, *l_ring = NULL, *pcap_ring=NULL;
		int dlt = pcap_datalink(pcap_file);
		while((pcap_buf = pcap_next(pcap_file, &pkt_header)) != NULL){
			// pkt_header.caplen
			register int pktlen = pkt_header.len;
			if(dlt == DLT_RAW) pktlen+= sizeof(eth_hdr);
			c_ring = malloc(pktlen+sizeof(struct pcap_ring_t));
			c_ring->len = pktlen;
			c_ring->next = NULL;
			if(dlt == DLT_RAW){
				memcpy(c_ring->buf, &eth_hdr, sizeof(eth_hdr));
				memcpy(c_ring->buf+sizeof(eth_hdr), pcap_buf, pkt_header.caplen);
			}else{
				memcpy(c_ring->buf, pcap_buf, pkt_header.caplen);
				memcpy(c_ring->buf, &eth_hdr, 12);
			}
			if(pcap_ring == NULL) {
				pcap_ring = c_ring;
			} else {
				l_ring->next = c_ring;
			}
			l_ring = c_ring;
		}
		l_ring->next = pcap_ring;
		pcap_close(pcap_file);
		return pcap_ring;
}

struct stat_out {
	int	x,y;
	char	*name, *format;
	int	val_len;
};

struct stat_out StatArray[] = {{0,0, "Loop", "%14llu", 14}, {0,0,"PPS","%8lu",8},
			       {0,0,"L2 Kbps", "%8lu",8}, {0,0,"L3 Kbps", "%8lu",8}, {0,0,"L1 Kbps", "%8lu",8},
			       {0,0,"Total bytes", "%12lu",12}, {0,0,"Total pkt","%8lu",8},
			       {0,0,"Sample bytes", "%12lu",12}, {0,0,"Sample pkt","%8lu",8}
//			       {0,0,"Sample L2 Kbps", "%8lu",8}, {0,0,"Sample PPS","%8lu",8}
};
				       
int
main(int argc, char **argv)
{
	int ch;
	int burst = 16;
	char *ifname = NULL, *samplefile=NULL, *progname = argv[0];

	size_t sfreq=sizeof(freq);
	sysctlbyname("machdep.tsc_freq", &freq, &sfreq, (void *)NULL, 0);
	printf("Freq %" PRIu64 "\n",freq);
	while ( (ch = getopt(argc, argv,"s:i:b:l:p:1:")) != -1) {
		switch(ch) {
		default:
			fprintf(stderr, "bad option %c %s", ch, optarg);
			usage(progname);
			break;
		case 's':
			Speed = atol(optarg);
			break;
		case 'i':       /* interface */
			ifname = optarg;
			break;
		case 'l':
			LoopCnt = atol(optarg);
			break;
		case 'p':
			SendMode = atoi(optarg);
			break;
		case 'b':       /* burst */
			burst = atoi(optarg);
			break;
		case '1':	/* one loop file */
			samplefile = optarg;
			break;
		}
	}
	argv += optind;
	argc -= optind;
	
	if (ifname == NULL) {
		fprintf(stderr, "missing ifname");
		usage(progname);
	}
	struct targ targs[1];
	bzero(&targs[0],  sizeof(targs[0]));
	targs[0].me = 0;
	targs[0].nfiles=argc;
	targs[0].batch_size = burst;
	targs[0].SendMode = SendMode;
	targs[0].LoopCnt = LoopCnt;
	if(samplefile) targs[0].nsamplefiles=1;
	struct nmreq nmr;
	bzero(&nmr, sizeof(nmr));
	nmr.nr_version = NETMAP_API;
	int fd = open("/dev/netmap", O_RDWR);
	int devqueues = 1;      /* how many device queues */
	if (fd == -1) {
		fprintf(stderr, "Unable to open /dev/netmap\n");
		// fail later
	} else {
		if ((ioctl(fd, NIOCGINFO, &nmr)) == -1) {
			fprintf(stderr, "Unable to get if info without name");
		} else {
			fprintf(stderr, "map size is %d Kb", nmr.nr_memsize >> 10);
		}
		bzero(&nmr, sizeof(nmr));
		nmr.nr_version = NETMAP_API;
		strncpy(nmr.nr_name, ifname, sizeof(nmr.nr_name));
		if ((ioctl(fd, NIOCGINFO, &nmr)) == -1) {
			fprintf(stderr, "Unable to get if info for %s", ifname);
		}
		fprintf(stdout, "ifname: numrings %d\n", nmr.nr_tx_rings);
		devqueues = nmr.nr_tx_rings;
	}
	fprintf(stderr, "mmapping %d Kbytes", nmr.nr_memsize>>10);
	void *mmap_addr;                /* the mmap address */
	mmap_addr = (struct netmap_d *) mmap(0, nmr.nr_memsize, PROT_WRITE | PROT_READ, MAP_SHARED, fd, 0);
	if (mmap_addr == MAP_FAILED) {
		fprintf(stderr, "Unable to mmap %d KB", nmr.nr_memsize >> 10);
		// continue, fail later
	}
	if (ioctl(fd, NIOCREGIF, &nmr) == -1) {
		fprintf(stderr, "Unable to register interface %s", ifname);
		//continue, fail later
	}
	/* Print some debug information. */
	fprintf(stdout, "%s: %d queues.\n", ifname, devqueues);
	/* Exit if something went wrong. */
	if (fd < 0) {
		fprintf(stderr, "aborting");
		usage(progname);
	}

	pcap_ring = calloc(targs[0].nfiles,sizeof(*pcap_ring));
	pcap_sample_ring = calloc(targs[0].nsamplefiles,sizeof(*pcap_sample_ring));
	for(int i=0; i<targs[0].nfiles; i++){
		pcap_ring[i] = load_pcap_file(argv[i]);
	}
	for(int i=0; i<targs[0].nsamplefiles; i++){
		pcap_sample_ring[i] = load_pcap_file(samplefile);
	}
	if(pcap_ring[0] == NULL) {
		fprintf(stderr, "No files to send\n");
		exit(1);
	}
	
	struct netmap_if *tnifp;
	struct nmreq tifreq;
	int tfd;
	/* register interface. */
	tfd = open("/dev/netmap", O_RDWR);
	if (tfd == -1) {
		fprintf(stderr, "Unable to open /dev/netmap");
		return(-1); //continue;
	}
	bzero(&tifreq, sizeof(tifreq));
	tifreq.nr_version = NETMAP_API;
	strncpy(tifreq.nr_name, ifname, sizeof(tifreq.nr_name));
	tifreq.nr_ringid = 0;
	if ((ioctl(tfd, NIOCREGIF, &tifreq)) == -1) {
		fprintf(stderr, "Unable to register %s", ifname);
		return(-1); // continue;
	}
	tnifp = NETMAP_IF(mmap_addr, tifreq.nr_offset);
	targs[0].nmr = tifreq;
	targs[0].nifp = tnifp;
	targs[0].fd = tfd;

	sleep(3);
	/* ncurses */
	initscr(); 
	/*
	  cbreak(); noecho();
	  nonl();
	  intrflush(stdscr, FALSE);
	  keypad(stdscr, TRUE);
	*/
	int maxx, maxy;
	getmaxyx(stdscr, maxy, maxx);
	int Log_h = maxy/3;
	WINDOW *WIN_Log = newwin(Log_h, maxx, maxy-Log_h, 0);
	scrollok(WIN_Log, 1);
	wsetscrreg(WIN_Log, 1, Log_h-1);
	WINDOW *WIN_Stat = newwin(maxy-Log_h, maxx, 0, 0);
	clear();
	char buf[1024];
	wmove(WIN_Stat, 0, 0);
	whline(WIN_Stat, ACS_HLINE, maxx);
	snprintf(buf,1024," Interface %s statistic ", ifname);
	mvwaddstr(WIN_Stat, 0, (maxx-strlen(buf))/2, buf);
	int line=1, col=0;
	for(unsigned i=0; i<sizeof(StatArray)/sizeof(StatArray[0]); i++){
		int len = strlen(StatArray[i].name);
		if(len+col+1+StatArray[i].val_len > maxx){ line++; col=0; }
		mvwaddstr(WIN_Stat, line, col, StatArray[i].name);
		StatArray[i].x = col+len+1;
		StatArray[i].y = line;
		col += len+1+StatArray[i].val_len+1;
	}
	wrefresh(WIN_Stat);
	wmove(WIN_Log, 0, 0);
	whline(WIN_Log, ACS_HLINE, maxx);
	snprintf(buf,1024," Log ");
	mvwaddstr(WIN_Log, 0, (maxx-strlen(buf))/2, buf);
	wmove(WIN_Log, 1, 0);
	wrefresh(WIN_Log);
	cpuset_t cpuset;
	CPU_ZERO(&cpuset);
	CPU_SET(2, &cpuset);
	pthread_attr_t pattr;
	pthread_attr_init(&pattr);
	pthread_attr_setaffinity_np(&pattr, sizeof(cpuset), &cpuset);
	if (pthread_create(&targs[0].thread, &pattr, send_thread, &targs[0]) == -1) {
		D("Unable to create thread %d", targs[0].me);
	}

	struct timeval toc;
	uint64_t last_pkt = 0, last_bytes =0; //, last_sample_pkt = 0, last_sample_bytes = 0;
	gettimeofday(&toc, NULL);
	for (;;) {
		struct timeval now, delta;
		uint64_t pps, bps, dt;
		delta.tv_sec = 1; delta.tv_usec = 0;
		select(0, NULL, NULL, NULL, &delta);
		gettimeofday(&now, NULL);
		timersub(&now, &toc, &toc);
		pps = targs[0].p_count-last_pkt;
		bps = targs[0].p_bytes-last_bytes;
		dt = toc.tv_sec* 1000000 + toc.tv_usec;
		int i = 0;
		mvwprintw(WIN_Stat, StatArray[i].y, StatArray[i].x, StatArray[i].format, targs[0].run_count); i++;
		mvwprintw(WIN_Stat, StatArray[i].y, StatArray[i].x, StatArray[i].format, (long)(pps*1000000/dt)); i++;
		mvwprintw(WIN_Stat, StatArray[i].y, StatArray[i].x, StatArray[i].format, (long)(bps*8000/dt)); i++;
		mvwprintw(WIN_Stat, StatArray[i].y, StatArray[i].x, StatArray[i].format, (long)((bps - pps*18)*8000/dt)); i++;
		mvwprintw(WIN_Stat, StatArray[i].y, StatArray[i].x, StatArray[i].format, (long)((bps+pps*(8+12))*8000/dt)); i++;
		mvwprintw(WIN_Stat, StatArray[i].y, StatArray[i].x, StatArray[i].format, (long)(targs[0].p_bytes)); i++;
		mvwprintw(WIN_Stat, StatArray[i].y, StatArray[i].x, StatArray[i].format, (long)(targs[0].p_count)); i++;
		mvwprintw(WIN_Stat, StatArray[i].y, StatArray[i].x, StatArray[i].format, (long)(targs[0].sample_bytes)); i++;
		mvwprintw(WIN_Stat, StatArray[i].y, StatArray[i].x, StatArray[i].format, (long)(targs[0].sample_count)); i++;
#if 0
		uint64_t sam_pps, sam_bps;
		sam_pps = targs[0].sample_count - last_sample_pkt;
		sam_bps = targs[0].sample_bytes - last_sample_bytes;
		last_sample_pkt = targs[0].sample_count; last_sample_bytes = targs[0].sample_bytes;
		mvwprintw(WIN_Stat, StatArray[i].y, StatArray[i].x, StatArray[i].format, (long)(sam_bps*8000/dt)); i++;
		mvwprintw(WIN_Stat, StatArray[i].y, StatArray[i].x, StatArray[i].format, (long)(sam_pps*1000000/dt)); i++;
#endif
		wrefresh(WIN_Stat);
		last_pkt = targs[0].p_count;
		last_bytes = targs[0].p_bytes;
		toc = now;
		if(targs[0].done) break;
	}
//	pthread_join(targs[0].thread, NULL);
	ioctl(fd, NIOCUNREGIF, &nmr);
	munmap(mmap_addr, nmr.nr_memsize);
	close(fd);

	return (0);
}
